/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include <gtest/gtest.h>

#include <faiss/utils/utils.h>

TEST(Threading, openmp) {
    EXPECT_TRUE(faiss::check_openmp());
}
